{% extends "iadmin/change_form.html" %}{% load geo %}
{% block content %}
    {{ block.super }}
    {% if nodes %}
        <ul>
            {% recursetree nodes %}
                <li>
                    {{ node.name }}
                    {% if not node.is_leaf_node %}
                        <ul class="children">
                            {{ children }}
                        </ul>
                    {% endif %}
                </li>
            {% endrecursetree %}
        </ul>
    {% endif %}
{% endblock content %}
